--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- DDR3 Controller for Xilinx Spartan6
-- http://www.ohwr.org/projects/ddr3-sp6-core
--------------------------------------------------------------------------------
--
-- unit name:   ddr3_ctrl_wb
--
-- description: Wishbone interface for DDR3 controller.
--
-- Important restrictions:
-- * pipelined wishbone only
-- * cannot mix read and writes during the same wishbone cycle
-- * must use consecutive addresses during the same wishbone cycle
-- * must drop CYC when done
--
--------------------------------------------------------------------------------
-- Copyright CERN 2011-2020
--------------------------------------------------------------------------------
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 2.0 (the "License"); you may not use this file except
-- in compliance with the License. You may obtain a copy of the License at
-- http://solderpad.org/licenses/SHL-2.0.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gencores_pkg.all;
use work.wishbone_pkg.all;
use work.genram_pkg.all;

entity ddr3_ctrl_wb is

  generic(
    -- DDR3 byte address width
    g_BYTE_ADDR_WIDTH  : integer                        := 30;
    -- Data mask size (8-bit granularity)
    g_MASK_SIZE        : integer                        := 4;
    -- Data width
    g_DATA_PORT_SIZE   : integer                        := 32;
    -- Granularity for wb_addr_i signal (BYTE|WORD)
    g_ADDR_GRANULARITY : t_wishbone_address_granularity := WORD
    );

  port(
    ----------------------------------------------------------------------------
    -- Reset input (active low)
    ----------------------------------------------------------------------------
    rst_n_i             : in  std_logic;
    ----------------------------------------------------------------------------
    -- DDR controller port
    ----------------------------------------------------------------------------
    ddr_cmd_clk_o       : out std_logic;
    ddr_cmd_en_o        : out std_logic;
    ddr_cmd_instr_o     : out std_logic_vector(2 downto 0);
    ddr_cmd_bl_o        : out std_logic_vector(5 downto 0);
    ddr_cmd_byte_addr_o : out std_logic_vector(g_BYTE_ADDR_WIDTH - 1 downto 0);
    ddr_cmd_empty_i     : in  std_logic;
    ddr_cmd_full_i      : in  std_logic;
    ddr_wr_clk_o        : out std_logic;
    ddr_wr_en_o         : out std_logic;
    ddr_wr_mask_o       : out std_logic_vector(g_MASK_SIZE - 1 downto 0);
    ddr_wr_data_o       : out std_logic_vector(g_DATA_PORT_SIZE - 1 downto 0);
    ddr_wr_full_i       : in  std_logic;
    ddr_wr_empty_i      : in  std_logic;
    ddr_wr_count_i      : in  std_logic_vector(6 downto 0);
    ddr_wr_underrun_i   : in  std_logic;
    ddr_wr_error_i      : in  std_logic;
    ddr_rd_clk_o        : out std_logic;
    ddr_rd_en_o         : out std_logic;
    ddr_rd_data_i       : in  std_logic_vector(g_DATA_PORT_SIZE - 1 downto 0);
    ddr_rd_full_i       : in  std_logic;
    ddr_rd_empty_i      : in  std_logic;
    ddr_rd_count_i      : in  std_logic_vector(6 downto 0);
    ddr_rd_overflow_i   : in  std_logic;
    ddr_rd_error_i      : in  std_logic;
    ----------------------------------------------------------------------------
    -- Wishbone bus port
    ----------------------------------------------------------------------------
    wb_rst_n_i          : in  std_logic;
    wb_clk_i            : in  std_logic;
    wb_sel_i            : in  std_logic_vector(g_MASK_SIZE - 1 downto 0);
    wb_cyc_i            : in  std_logic;
    wb_stb_i            : in  std_logic;
    wb_we_i             : in  std_logic;
    wb_addr_i           : in  std_logic_vector(31 downto 0);
    wb_data_i           : in  std_logic_vector(g_DATA_PORT_SIZE - 1 downto 0);
    wb_data_o           : out std_logic_vector(g_DATA_PORT_SIZE - 1 downto 0);
    wb_ack_o            : out std_logic;
    wb_stall_o          : out std_logic);

end entity ddr3_ctrl_wb;

architecture arch of ddr3_ctrl_wb is

  ------------------------------------------------------------------------------
  -- Constants and functions declaration
  ------------------------------------------------------------------------------

  -- Since the Command FIFO of the DDR is 4-deep and its Read/Write FIFOs are
  -- 64-deep, we set the burst length to 16, to allow four commands to span
  -- the full data FIFO.
  constant c_DDR_BURST_LEN : integer := 16;

  -- How many more commands to store locally. Having a small local CMD FIFO
  -- (along with a WR and RD FIFO) helps with reaching timing closure.
  constant c_CMD_FIFO_DEPTH : integer := 2;
  constant c_WR_FIFO_DEPTH  : integer := c_CMD_FIFO_DEPTH * c_DDR_BURST_LEN;
  constant c_RD_FIFO_DEPTH  : integer := c_CMD_FIFO_DEPTH * c_DDR_BURST_LEN;

  constant c_CMD_FIFO_WIDTH : integer := g_BYTE_ADDR_WIDTH + 9;
  constant c_WR_FIFO_WIDTH  : integer := g_DATA_PORT_SIZE + g_MASK_SIZE;
  constant c_RD_FIFO_WIDTH  : integer := g_DATA_PORT_SIZE;

  -- all command addresses need to be aligned to g_DATA_PORT_SIZE as per UG388.
  -- g_ADDR_GRANULARITY = WORD:
  --  32bit data: cmd_addr[29..2] <= wb_addr[27..0]
  --  64bit data: cmd_addr[29..3] <= wb_addr[27..1]
  -- 128bit data: cmd_addr[29..4] <= wb_addr[27..2]
  -- g_ADDR_GRANULARITY = BYTE:
  --  32bit data: cmd_addr[29..2] <= wb_addr[29..2]
  --  64bit data: cmd_addr[29..3] <= wb_addr[29..3]
  -- 128bit data: cmd_addr[29..4] <= wb_addr[29..4]
  function f_align_addr (constant addr : std_logic_vector) return std_logic_vector is
    variable l2ds : integer;
    variable ret  : std_logic_vector(g_BYTE_ADDR_WIDTH-1 downto 0);
  begin
    l2ds := log2_ceil(g_DATA_PORT_SIZE/8);
    ret  := (others => '0');
    if g_ADDR_GRANULARITY = WORD then
      ret(g_BYTE_ADDR_WIDTH-1 downto l2ds) :=
        addr(g_BYTE_ADDR_WIDTH-3 downto l2ds-2);
    else
      ret(g_BYTE_ADDR_WIDTH-1 downto l2ds) :=
        addr(g_BYTE_ADDR_WIDTH-1 downto l2ds);
    end if;
    return ret;
  end function f_align_addr;

  ------------------------------------------------------------------------------
  -- Signals and types declaration
  ------------------------------------------------------------------------------

  type t_cmd_fsm_state is (S_IDLE, S_BUILD_BURST, S_EXEC_BURST);

  signal cmd_fsm_state : t_cmd_fsm_state := S_IDLE;

  signal wb_stall     : std_logic := '0';
  signal wb_stb_valid : std_logic := '0';

  signal ddr_wr_en     : std_logic                    := '0';
  signal ddr_cmd_en    : std_logic                    := '0';
  signal ddr_cmd_instr : std_logic_vector(2 downto 0) := (others => '0');
  signal ddr_cmd_bl    : std_logic_vector(5 downto 0) := (others => '0');

  signal ddr_cmd_byte_addr : std_logic_vector(g_BYTE_ADDR_WIDTH - 1 downto 0) := (others => '0');

  signal cmd_fifo_din  : std_logic_vector(c_CMD_FIFO_WIDTH-1 downto 0) := (others => '0');
  signal cmd_fifo_dout : std_logic_vector(c_CMD_FIFO_WIDTH-1 downto 0) := (others => '0');

  signal cmd_fifo_we    : std_logic := '0';
  signal cmd_fifo_rd    : std_logic := '0';
  signal cmd_fifo_full  : std_logic := '0';
  signal cmd_fifo_empty : std_logic := '0';

  signal cmd_fifo_rd_disable : std_logic := '0';

  signal ddr_cmd_en_out : std_logic := '0';

  signal wr_fifo_din  : std_logic_vector(c_WR_FIFO_WIDTH-1 downto 0) := (others => '0');
  signal wr_fifo_dout : std_logic_vector(c_WR_FIFO_WIDTH-1 downto 0) := (others => '0');

  signal wr_fifo_we    : std_logic := '0';
  signal wr_fifo_rd    : std_logic := '0';
  signal wr_fifo_full  : std_logic := '0';
  signal wr_fifo_empty : std_logic := '0';

  signal rd_fifo_din  : std_logic_vector(c_RD_FIFO_WIDTH-1 downto 0) := (others => '0');
  signal rd_fifo_dout : std_logic_vector(c_RD_FIFO_WIDTH-1 downto 0) := (others => '0');

  signal rd_fifo_we    : std_logic := '0';
  signal rd_fifo_rd    : std_logic := '0';
  signal rd_fifo_full  : std_logic := '0';
  signal rd_fifo_empty : std_logic := '0';

begin

  ------------------------------------------------------------------------------
  -- Sanity checks
  ------------------------------------------------------------------------------
  -- Only port sizes of 32, 64, and 128 are supported, as per UG388.
  assert (g_DATA_PORT_SIZE = 32 or g_DATA_PORT_SIZE = 64 or g_DATA_PORT_SIZE = 128)
    report "Provided data port size must be 32, 64 or 128" severity FAILURE;

  -- Command path assertions
  assert unsigned(ddr_cmd_byte_addr(f_log2_ceil(g_DATA_PORT_SIZE/8)-1 downto 0)) = 0
    report "CMD byte address to data port size misalignment" severity FAILURE;

  -- Write datapath assertions
  assert ddr_wr_underrun_i /= '1'
    report "WR FIFO underrun" severity FAILURE;
  assert ddr_wr_error_i /= '1'
    report "WR FIFO error" severity FAILURE;

  -- Read datapath assertions
  assert ddr_rd_overflow_i /= '1'
    report "RD FIFO overflow" severity FAILURE;
  assert ddr_rd_error_i /= '1'
    report "RD FIFO error" severity FAILURE;

  ------------------------------------------------------------------------------
  -- CMD, WR and RD internal FIFOs
  --
  -- These are used to help with reaching timing closure with slow signals
  -- such as ddr_cmd_full, by decoupling them from the rest of the FPGA.
  ------------------------------------------------------------------------------

  --
  -- CMD FIFO and control
  --

  cmp_cmd_fifo : entity work.generic_sync_fifo
    generic map (
      g_DATA_WIDTH             => c_CMD_FIFO_WIDTH,
      g_SIZE                   => c_CMD_FIFO_DEPTH,
      g_SHOW_AHEAD             => TRUE,
      g_SHOW_AHEAD_LEGACY_MODE => FALSE)
    port map (
      rst_n_i => wb_rst_n_i,
      clk_i   => wb_clk_i,
      d_i     => cmd_fifo_din,
      we_i    => cmd_fifo_we,
      q_o     => cmd_fifo_dout,
      rd_i    => cmd_fifo_rd,
      empty_o => cmd_fifo_empty,
      full_o  => cmd_fifo_full);

  cmd_fifo_din <= ddr_cmd_byte_addr & ddr_cmd_instr & ddr_cmd_bl;
  cmd_fifo_we  <= ddr_cmd_en;

  ddr_cmd_en_out <= cmd_fifo_rd;

  ddr_cmd_bl_o        <= cmd_fifo_dout(5 downto 0);
  ddr_cmd_instr_o     <= cmd_fifo_dout(8 downto 6);
  ddr_cmd_byte_addr_o <= cmd_fifo_dout(c_CMD_FIFO_WIDTH-1 downto 9);
  ddr_cmd_en_o        <= ddr_cmd_en_out;
  ddr_cmd_clk_o       <= wb_clk_i;

  -- We need to make sure that there is always a delay between consecutive commands,
  -- otherwise the cmp_cmd_fifo above can queue up several write commands, and
  -- release them immediately one after the other, causing the MCB's WR FIFO to
  -- underrun.
  p_cmd_fifo_rd_monostable : process (wb_clk_i) is
    variable cmd_fifo_rd_timer : unsigned(5 downto 0) := (others => '0');
  begin
    if rising_edge(wb_clk_i) then
      if wb_rst_n_i = '0' then
        cmd_fifo_rd_disable <= '0';
      else
        if cmd_fifo_rd = '1' then
          cmd_fifo_rd_timer   := to_unsigned(c_DDR_BURST_LEN, cmd_fifo_rd_timer'length);
          cmd_fifo_rd_disable <= '1';
        elsif cmd_fifo_rd_disable <= '1' then
          cmd_fifo_rd_timer    := cmd_fifo_rd_timer - 1;
          if cmd_fifo_rd_timer = 0 then
            cmd_fifo_rd_disable <= '0';
          end if;
        end if;
      end if;
    end if;
  end process p_cmd_fifo_rd_monostable;

  p_cmd_fifo_rd : process (cmd_fifo_dout(6 downto 0), cmd_fifo_empty,
                           cmd_fifo_rd_disable, ddr_cmd_full_i,
                           ddr_wr_count_i) is
    variable next_cmd_len : unsigned(6 downto 0) := (others => '0');
    variable wr_fifo_cnt  : unsigned(6 downto 0) := (others => '0');
  begin
    cmd_fifo_rd <= '0';
    if cmd_fifo_empty = '0' and ddr_cmd_full_i = '0' and cmd_fifo_rd_disable = '0' then
      next_cmd_len := unsigned('0' & cmd_fifo_dout(5 downto 0));
      wr_fifo_cnt  := unsigned(ddr_wr_count_i);
      if cmd_fifo_dout(6) = '1' or wr_fifo_cnt > next_cmd_len then
        cmd_fifo_rd <= '1';
      end if;
    end if;
  end process p_cmd_fifo_rd;

  --
  -- WR FIFO and control
  --

  cmp_wr_fifo : entity work.generic_sync_fifo
    generic map (
      g_DATA_WIDTH             => c_WR_FIFO_WIDTH,
      g_SIZE                   => c_WR_FIFO_DEPTH,
      g_SHOW_AHEAD             => TRUE,
      g_SHOW_AHEAD_LEGACY_MODE => FALSE)
    port map (
      rst_n_i => wb_rst_n_i,
      clk_i   => wb_clk_i,
      d_i     => wr_fifo_din,
      we_i    => wr_fifo_we,
      q_o     => wr_fifo_dout,
      rd_i    => wr_fifo_rd,
      empty_o => wr_fifo_empty,
      full_o  => wr_fifo_full);

  -- Write to the Write FIFO of the DDR if there is a valid WB write cycle
  -- and we are not stalling (in which case we should not write the same data
  -- multiple times).
  wr_fifo_we  <= wb_stb_valid and wb_we_i and not (wr_fifo_full or cmd_fifo_full);
  wr_fifo_din <= (not wb_sel_i) & wb_data_i;

  wr_fifo_rd <= not (wr_fifo_empty or ddr_wr_full_i);
  ddr_wr_en  <= wr_fifo_rd;

  ddr_wr_mask_o <= wr_fifo_dout(c_WR_FIFO_WIDTH-1 downto g_DATA_PORT_SIZE);
  ddr_wr_data_o <= wr_fifo_dout(g_DATA_PORT_SIZE-1 downto 0);
  ddr_wr_en_o   <= ddr_wr_en;
  ddr_wr_clk_o  <= wb_clk_i;

  --
  -- RD FIFO and control
  --

  cmp_rd_fifo : entity work.generic_sync_fifo
    generic map (
      g_DATA_WIDTH => c_RD_FIFO_WIDTH,
      g_SIZE       => c_RD_FIFO_DEPTH,
      g_SHOW_AHEAD => FALSE)
    port map (
      rst_n_i => wb_rst_n_i,
      clk_i   => wb_clk_i,
      d_i     => rd_fifo_din,
      we_i    => rd_fifo_we,
      q_o     => rd_fifo_dout,
      rd_i    => rd_fifo_rd,
      empty_o => rd_fifo_empty,
      full_o  => rd_fifo_full);

  rd_fifo_din <= ddr_rd_data_i;
  rd_fifo_we  <= not (ddr_rd_empty_i or rd_fifo_full);

  -- If there is data to be read from the Read FIFO of the DDR, it is safe to
  -- assume that it is the result of a valid WB read cycle earlier. As long as
  -- the WB cycle is still ongoing, we should output the data, even if currently
  -- the slave is stalling.
  rd_fifo_rd <= wb_cyc_i and not rd_fifo_empty;

  ddr_rd_en_o  <= rd_fifo_we;
  ddr_rd_clk_o <= wb_clk_i;

  ------------------------------------------------------------------------------
  -- Pipelined WB slave
  ------------------------------------------------------------------------------

  wb_stb_valid <= wb_cyc_i and wb_stb_i;

  -- The slave will stall under any of the two following conditions:
  -- 1. When the DDR Command FIFO is full and there is an active WB cycle.
  -- 2. When the DDR Write FIFO is full and there is an active WB *write* cycle.
  wb_stall <= wb_stb_valid and (cmd_fifo_full or (wb_we_i and wr_fifo_full));

  -- Whenever we read/write something from/to the Read/Write FIFOs
  -- of the DDR we assert ACK.
  p_wb_ack : process (wb_clk_i) is
  begin
    if rising_edge(wb_clk_i) then
      wb_ack_o <= wr_fifo_we or rd_fifo_rd;
    end if;
  end process p_wb_ack;

  wb_data_o  <= rd_fifo_dout;
  wb_stall_o <= wb_stall;

  ------------------------------------------------------------------------------
  -- DDR Command FIFO control FSM
  ------------------------------------------------------------------------------

  -- This FSM controls the generation of the 'cmd_byte_addr', 'cmd_en',
  -- 'cmd_instr' and 'cmd_bl' signals.
  p_cmd_fsm : process (wb_clk_i) is
    variable ddr_burst_cnt  : unsigned(5 downto 0)          := (others => '0');
    variable next_cmd_instr : std_logic := '0';
    variable next_cmd_addr  : std_logic_vector(31 downto 0) := (others => '0');

  begin
    if rising_edge(wb_clk_i) then
      if wb_rst_n_i = '0' then
        cmd_fsm_state     <= S_IDLE;
        ddr_cmd_instr     <= (others => '0');
        ddr_cmd_byte_addr <= (others => '0');
        ddr_cmd_bl        <= (others => '0');
        ddr_burst_cnt     := (others => '0');
      else

        case cmd_fsm_state is

          when S_IDLE =>
            ddr_cmd_en <= '0';
            if wb_stb_valid = '1' and wb_stall = '0' then
              ddr_burst_cnt  := to_unsigned(1, 6);
              next_cmd_instr := not wb_we_i;
              next_cmd_addr  := wb_addr_i;
              cmd_fsm_state  <= S_BUILD_BURST;
            end if;

          when S_BUILD_BURST =>
            ddr_cmd_en <= '0';
            if wb_stb_valid = '1' and wb_stall = '0' then
              ddr_burst_cnt := ddr_burst_cnt + 1;
            end if;
            -- Bursts are dispathced when:
            -- 1: We reach the burst length limit.
            -- 2: We are in a read cycle and STB is de-asserted.
            -- 3: We are in a write cycle and CYC is de-asserted.
            if (ddr_burst_cnt = to_unsigned(c_DDR_BURST_LEN, 6) or
                (wb_stb_valid = '0' and next_cmd_instr = '1') or
                (wb_cyc_i = '0' and next_cmd_instr = '0')) then
              ddr_cmd_byte_addr <= f_align_addr(next_cmd_addr);
              ddr_cmd_instr(0)  <= next_cmd_instr;
              if (ddr_burst_cnt = to_unsigned(c_DDR_BURST_LEN, 6)) then
                ddr_cmd_bl <= "001111";
              else
                ddr_cmd_bl <= std_logic_vector(ddr_burst_cnt - 1);
              end if;
              cmd_fsm_state <= S_EXEC_BURST;
            end if;

          when S_EXEC_BURST =>
            if cmd_fifo_full = '0' then
              ddr_cmd_en <= '1';
              -- Check if we need to start another burst or not.
              if wb_stb_valid = '1' and wb_stall = '0' then
                ddr_burst_cnt  := to_unsigned(1, 6);
                next_cmd_instr := not wb_we_i;
                next_cmd_addr  := wb_addr_i;
                cmd_fsm_state  <= S_BUILD_BURST;
              else
                cmd_fsm_state <= S_IDLE;
              end if;
            end if;

          when others =>
            cmd_fsm_state <= S_IDLE;

        end case;
      end if;
    end if;
  end process p_cmd_fsm;

end architecture arch;
